from path import Path
from subprocess import call

from charmhelpers.core import hookenv
from charmhelpers.core import unitdata
from charmhelpers.core.charmframework.helpers import any_ready_unit
from jujubigdata import utils
from jujubigdata.relations import Kafka


# Extended status support
# We call update_blocked_status from the "requires" section of our service
# block, so be sure to return True. Otherwise, we'll block the "requires"
# and never move on to callbacks. The other status update methods are called
# from the "callbacks" section and therefore don't need to return True.
def update_blocked_status():
    # Report blocked if missing config, regardless of our charm.active flag
    if not hookenv.config()['kafka_topic']:
        hookenv.status_set('blocked', 'Missing kafka_topic config value')
        return True
    elif not hookenv.config()['pubnub_channel']:
        hookenv.status_set('blocked', 'Missing pubnub_channel config value')
        return True
    elif not hookenv.config()['pubnub_key']:
        hookenv.status_set('blocked', 'Missing pubnub_key config value')
        return True

    # If we're active, there's no need to test relations
    if unitdata.kv().get('charm.active', False):
        return True

    # If we've made it this far, we've got config but aren't yet active.
    # Make sure we've got required (and ready) relations.
    if not Kafka().connected_units():
        hookenv.status_set('blocked', 'Waiting for relation to apache-kafka')
    elif not Kafka().is_ready():
        hookenv.status_set('waiting', 'Waiting for Kafka to become ready')
    return True


def update_working_status():
    if unitdata.kv().get('charm.active', False):
        hookenv.status_set('maintenance', 'Updating configuration')
        return
    hookenv.status_set('maintenance', 'Setting up Kafka-PubNub')


def update_active_status():
    unitdata.kv().set('charm.active', True)
    # The charm may be active from a 'relation' standpoint, but report blocked
    # if we're missing required config.
    if not hookenv.config()['kafka_topic']:
        hookenv.status_set('blocked', 'Missing kafka_topic config value')
    elif not hookenv.config()['pubnub_channel']:
        hookenv.status_set('blocked', 'Missing pubnub_channel config value')
    elif not hookenv.config()['pubnub_key']:
        hookenv.status_set('blocked', 'Missing pubnub_key config value')
    else:
        hookenv.status_set('active', 'Ready')


def clear_active_flag():
    unitdata.kv().set('charm.active', False)


# Main Kafka-PubNub class for callbacks
class KafkaPubNub(object):
    def __init__(self, dist_config):
        self.dist_config = dist_config
        self.resources = {
        }
        self.verify_resources = utils.verify_resources(*self.resources.values())

    def is_installed(self):
        return unitdata.kv().get('kafka_pubnub.installed')

    def install(self, force=False):
        if not force and self.is_installed():
            return
        self.copy_script()
        unitdata.kv().set('kafka_pubnub.installed', True)

    def copy_script(self):
        '''
        Copy charm version of kafka_pubnub.py to /home/ubuntu. This will
        happen at installation, and again if our prereqs ever go away
        (eg, the relation to apache-kafka).
        '''
        script_source = 'scripts/kafka_pubnub.py'
        script_target = '/home/ubuntu/kafka_pubnub.py'
        Path(script_source).copyfile(script_target)
        Path(script_target).chmod(0o755)
        Path(script_target).chown('ubuntu')

    def configure_pubnub(self):
        '''
        Set appropriate Kafka and Pubnub configuration in kafka_pubnub.py.
        '''
        pubnub_script = '/home/ubuntu/kafka_pubnub.py'

        # check for required config and relations
        config = hookenv.config()
        topic = config['kafka_topic']
        key = config['pubnub_key']
        channel = config['pubnub_channel']
        if (topic and key and channel and Kafka().connected_units()):
            unit, data = any_ready_unit('kafka')
            kafka_ip = utils.resolve_private_address(data['private-address'])
            kafka_port = data['port']

            utils.re_edit_in_place(pubnub_script, {
                r'JUJU_KAFKA_IP': '{}'.format(kafka_ip),
                r'JUJU_KAFKA_PORT': '{}'.format(kafka_port),
                r'JUJU_KAFKA_TOPIC': '{}'.format(topic),
                r'JUJU_PUBNUB_KEY': '{}'.format(key),
                r'JUJU_PUBNUB_CHANNEL': '{}'.format(channel),
            })
        else:
            self.cleanup()

    def cleanup(self):
        '''
        We get here when required relations or config are missing. Kill any
        kafka_pubnub.py scripts and reset it to the pristine charm source
        version so it can be re-configured once reqs are met.
        '''
        # TODO: stop any running kafka_pubnub.py PIDs
        # reset script to the unconfigured one from the charm
        self.copy_script()
